#![doc = include_str!("../README.md")]
#![warn(missing_docs)]

use std::{collections::HashMap, ffi::c_void, rc::Rc};

#[allow(unused)]
#[allow(non_snake_case)]
#[allow(non_camel_case_types)]
#[allow(non_upper_case_globals)]
#[allow(clippy::upper_case_acronyms)]
mod bindings {
    include!(concat!(env!("OUT_DIR"), "/schrift_bindings.rs"));
}

/// Retrieve the libschrift version information string.
pub fn version_info() -> &'static str {
    unsafe {
        std::ffi::CStr::from_ptr(bindings::sft_version())
            .to_str()
            .unwrap()
    }
}

/// Error type for schrift-rs operations.
#[derive(Debug)]
pub enum Error {
    /// Logic error in invocation of schrift-rs.
    LogicError(&'static str),
    /// Internal libschrift error from a malformed font file.
    FormatError(&'static str),
    /// I/O error while loading font.
    IOError(std::io::Error),
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        std::fmt::Debug::fmt(self, f)
    }
}

impl std::error::Error for Error { }

/// Size-independent font information, loaded from a single TTF file.
pub struct Font {
    _data: Box<[u8]>,
    sft_font: *mut bindings::SFT_Font,
}

impl Font {
    /// Load font information from a filesystem path.
    pub fn load_from_file(path: impl AsRef<std::path::Path>) -> Result<Rc<Self>, Error> {
        let font_data = std::fs::read(path.as_ref()).map_err(Error::IOError)?;
        let data = font_data.into_boxed_slice();

        let sft_font = unsafe { bindings::sft_loadmem(data.as_ptr() as *const c_void, data.len()) };
        if sft_font.is_null() {
            return Err(Error::FormatError("could not load font"));
        }

        Ok(Self {
            _data: data,
            sft_font,
        }
        .into())
    }
}

impl Drop for Font {
    fn drop(&mut self) {
        unsafe {
            bindings::sft_freefont(self.sft_font);
        }
    }
}

/// Line metrics for a [`Schrift`] instance.
#[derive(Debug)]
pub struct LineMetrics {
    /// Line ascender (height above baseline)
    pub ascender: f64,
    /// Line descender, as a negative value (height below baseline)
    pub descender: f64,
    /// Linegap (recommended spacing between lines). Use (ascender + descender + line_gap) for
    /// baseline-to-baseline distance.
    pub line_gap: f64,
}

/// Glyph metrics from a [`Schrift`] instance.
#[derive(Debug)]
pub struct GlyphMetrics {
    /// Horizontal advancing distance for this glyph.
    pub advance: f64,
    /// Left-bearing for this glyph, or distance between origin and min-x of the glyph's bounding
    /// box.
    pub left_bearing: f64,
    /// Pixel rendering y-offset from baseline.
    pub y_offset: i32,
    /// Minimum width for a pixel rendering of this glyph.
    pub min_width: i32,
    /// Minimum height for a pixel rendering of this glyph.
    pub min_height: i32,
}

/// TrueType glyph index type.
#[derive(Clone, Copy, PartialEq, Eq, Hash)]
pub struct Glyph(bindings::SFT_Glyph);

impl std::fmt::Debug for Glyph {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("Glyph").field(&self.0).finish()
    }
}

/// Glyph-pair kerning information.
#[derive(Debug)]
pub struct Kerning {
    /// Additional x-shift on top of usual glyph spacing information.
    pub x_shift: f64,
    /// Additional y-shift on top of usual glyph spacing information.
    pub y_shift: f64,
}

/// Size-dependent font functionality.
pub struct Schrift {
    // keep a reference to Font to keep the sft_font pointer in sft valid
    _font: Rc<Font>,
    sft: bindings::SFT,
}

impl Schrift {
    /// Instantiate a [`SchriftBuilder`] with a given font.
    pub fn build(font: Rc<Font>) -> SchriftBuilder {
        SchriftBuilder::new(font)
    }

    /// Compute the line metrics for this font context.
    pub fn line_metrics(&self) -> Result<LineMetrics, Error> {
        unsafe {
            let mut lm = bindings::SFT_LMetrics {
                ascender: 0.,
                descender: 0.,
                lineGap: 0.,
            };
            if bindings::sft_lmetrics(&self.sft as *const _, &mut lm as *mut _) != 0 {
                Err(Error::FormatError("line metrics table"))
            } else {
                Ok(LineMetrics {
                    ascender: lm.ascender,
                    descender: lm.descender,
                    line_gap: lm.lineGap,
                })
            }
        }
    }

    /// Perform unicode-to-glyph lookup.
    pub fn glyph_lookup(&self, c: char) -> Result<Glyph, Error> {
        unsafe {
            let mut g = bindings::SFT_Glyph::default();
            if bindings::sft_lookup(&self.sft as *const _, c as u32, &mut g as *mut _) != 0 {
                Err(Error::FormatError("glyph cmap"))
            } else {
                Ok(Glyph(g))
            }
        }
    }

    /// Look up glyph metrics for a given glyph index.
    pub fn glyph_metrics(&self, glyph: Glyph) -> Result<GlyphMetrics, Error> {
        unsafe {
            let mut gmet = std::mem::MaybeUninit::<bindings::SFT_GMetrics>::uninit();
            if bindings::sft_gmetrics(&self.sft as *const _, glyph.0, gmet.as_mut_ptr()) != 0 {
                Err(Error::FormatError("glyph mtx"))
            } else {
                let gmet = gmet.assume_init();
                Ok(GlyphMetrics {
                    advance: gmet.advanceWidth,
                    left_bearing: gmet.leftSideBearing,
                    y_offset: gmet.yOffset,
                    min_width: gmet.minWidth,
                    min_height: gmet.minHeight,
                })
            }
        }
    }

    /// Look up kerning information for a given pair of glyphs.
    pub fn kerning(&self, left: Glyph, right: Glyph) -> Result<Kerning, Error> {
        unsafe {
            let mut kern = std::mem::MaybeUninit::<bindings::SFT_Kerning>::uninit();
            if bindings::sft_kerning(&self.sft as *const _, left.0, right.0, kern.as_mut_ptr()) != 0
            {
                Err(Error::FormatError("kerning table"))
            } else {
                let kern = kern.assume_init();
                Ok(Kerning {
                    x_shift: kern.xShift,
                    y_shift: kern.yShift,
                })
            }
        }
    }

    /// Render a glyph into an alpha-valued bitmap.
    ///
    /// Note that 0=no colour, 255=full colour. Subpixel hinting is out of scope for libschrift.
    pub fn render(
        &self,
        glyph: Glyph,
        alpha_map: &mut [u8],
        width: usize,
        height: usize,
    ) -> Result<(), Error> {
        if alpha_map.len() < width * height {
            return Err(Error::LogicError(
                "alpha_map not large enough for given width and height",
            ));
        }

        unsafe {
            let img = bindings::SFT_Image {
                pixels: alpha_map.as_mut_ptr() as *mut c_void,
                width: width as i32,
                height: height as i32,
            };
            if bindings::sft_render(&self.sft as *const _, glyph.0, img) != 0 {
                Err(Error::FormatError("rendering failed"))
            } else {
                Ok(())
            }
        }
    }
}

/// Builder pattern for a [`Schrift`] instance.
pub struct SchriftBuilder {
    font: Rc<Font>,
    x_scale: Option<f64>,
    y_scale: Option<f64>,
    x_offset: Option<f64>,
    y_offset: Option<f64>,
    downward_y: Option<bool>,
}

impl SchriftBuilder {
    fn new(font: Rc<Font>) -> Self {
        Self {
            font,
            x_scale: None,
            y_scale: None,
            x_offset: None,
            y_offset: None,
            downward_y: None,
        }
    }

    /// Set both x- and y- scaling information.
    pub fn with_scale(mut self, scale: f64) -> Self {
        self.x_scale = Some(scale);
        self.y_scale = Some(scale);
        self
    }
    /// Set render offset information.
    pub fn with_offset(mut self, x: f64, y: f64) -> Self {
        self.x_offset = Some(x);
        self.y_offset = Some(y);
        self
    }
    /// Make the resulting coordinate space consider +y to be the bottom of the glyph, instead of
    /// the top.
    pub fn flag_y_downwards(mut self) -> Self {
        self.downward_y = Some(true);
        self
    }

    /// Construct a [`Schrift`] instance.
    pub fn build(self) -> Schrift {
        let scales = match (self.x_scale, self.y_scale) {
            (Some(xs), Some(ys)) => (xs, ys),
            (Some(xs), None) => (xs, xs),
            (None, Some(ys)) => (ys, ys),
            (None, None) => (12.0, 12.0),
        };

        Schrift {
            sft: bindings::SFT {
                font: self.font.sft_font,
                xScale: scales.0,
                yScale: scales.1,
                xOffset: self.x_offset.unwrap_or(0.0),
                yOffset: self.y_offset.unwrap_or(0.0),
                flags: if self.downward_y.unwrap_or(false) {
                    bindings::SFT_DOWNWARD_Y as i32
                } else {
                    0
                },
            },
            _font: self.font,
        }
    }
}

/// Caching wrapper around [`Schrift`] that treats format errors as panics.
pub struct CachedSchrift {
    schrift: Schrift,
    lmetrics: LineMetrics,
    glyph_cache: HashMap<char, Glyph>,
    image_cache: HashMap<Glyph, (Box<[u8]>, u32, u32)>,
    metric_cache: HashMap<Glyph, GlyphMetrics>,
    kern_cache: HashMap<(Glyph, Glyph), Kerning>,
}

impl CachedSchrift {
    /// Wrap a [`Schrift`] instance with this caching layer.
    pub fn new(schrift: Schrift) -> Self {
        Self {
            lmetrics: schrift.line_metrics().expect("couldn't get line metrics"),
            schrift,
            glyph_cache: Default::default(),
            image_cache: Default::default(),
            metric_cache: Default::default(),
            kern_cache: Default::default(),
        }
    }

    /// Get line metrics.
    pub fn line_metrics(&self) -> &LineMetrics {
        &self.lmetrics
    }

    /// Perform glyph lookup.
    pub fn glyph_lookup(&mut self, ch: char) -> Glyph {
        *self.glyph_cache.entry(ch).or_insert_with(|| {
            self.schrift
                .glyph_lookup(ch)
                .expect("couldn't lookup glyph")
        })
    }

    /// Get glyph metrics.
    pub fn glyph_metrics(&mut self, glyph: Glyph) -> &GlyphMetrics {
        self.metric_cache.entry(glyph).or_insert_with(|| {
            self.schrift
                .glyph_metrics(glyph)
                .expect("couldn't lookup glyph metrics")
        })
    }

    /// Get a rendered alphamap image for a given glyph.
    pub fn glyph_image(&mut self, glyph: Glyph) -> &(Box<[u8]>, u32, u32) {
        // this is written as an explicit contains_key()/get().unwrap() so that the long-lived
        // borrow happens inside the if block.
        if self.image_cache.contains_key(&glyph) {
            // this unwrap is always OK
            return self.image_cache.get(&glyph).unwrap();
        }
        let metrics = self.glyph_metrics(glyph);
        let w = metrics.min_width as usize;
        let h = metrics.min_height as usize;
        let mut data = vec![0u8; (metrics.min_width * metrics.min_height) as usize];
        self.schrift
            .render(glyph, &mut data, w, h)
            .expect("couldn't render glyph");

        self.image_cache
            .entry(glyph)
            .or_insert((data.into_boxed_slice(), w as u32, h as u32))
    }

    /// Get the kerning information for a given pair of glyphs.
    pub fn glyph_kerning(&mut self, left: Glyph, right: Glyph) -> &Kerning {
        self.kern_cache.entry((left, right)).or_insert_with(|| {
            self.schrift
                .kerning(left, right)
                .expect("couldn't get kerning")
        })
    }
}

#[cfg(test)]
mod test;
